//*****************************************************************************
//*
//*
//*		PlaneDrawer.cpp
//*
//*
//*****************************************************************************
//
//										start	= Erster  Pixel in der Scala (-1.0)
//										stop	= Letzter Pixel in der Scala (+1.0)
//										max		= Maximal Wert
//										min		= Minimal Wert
//
//	Umrechnung lineare Skala:			pos		= (val-offset)*fact
//										val		= pos*fact+offset;
//										fact	= (start-stop)/(max-min)    = 2/(max-min)
//										offset	= min-start/fact            = min+1/fact
//
//	Umrechnung logarithmische Skala:	pos		= log(val)*fact-offset
//										val		= exp((pos+offset)/fact)
//										fact	= (stop-start)/log(max/min)	= 2/log(max/min)
//										offset	= log(min)*fact+start       = log(min)*fact-1
//
#include	<math.h>
#include	<stdio.h>
#include	"StringFunctions.h"
#include	"MemFunctions.h"
#include	"PlaneDrawer.h"

#define 	GRID_MAX			256							// Maximale Teilung in X/Y
#define 	WHITE_DIF			128							// Maximaler Weisanteil fr Licht
#define 	WHITE_MAX			48							// Maximaler Weisanteil fr Licht
#define 	FRAME_XY			15							// X-Rand rund um Bild
#define 	FRAME_Z				15							// Y-Rand rund um Bild
#define 	FONT_SCALE			0.9							// Skalierunf fr Font-Pixel
#define 	FONT_HEIGHT_1		8							// Kleine Schrifthhe
#define 	FONT_HEIGHT_2		16							// Mitlere Schrifthhe
#define 	FONT_HEIGHT_3		36							// Groe Schrifthhe
#define 	FONT_OFFSET_1		7							// Kleiner Schrift-Offset zur Mitte
#define 	FONT_OFFSET_2		11							// Mittlerer Schrift-Offset zur Mitte
#define 	FONT_OFFSET_3		20							// Groer Schrift-Offset zur Mitte
#define 	FONT_WIDTH_1		6							// Breite eines kleinen   Buchstabens
#define 	FONT_WIDTH_2		10							// Breite eines mittleren Buchstabens
#define 	FONT_WIDTH_3		20							// Breite eines groen    Buchstabens
#define 	STEP_SIZE_NULL		0.08						// Bereite der Nullmarkierung
#define 	STEP_SIZE_MAIN		0.05						// Bereite der Textmarkierung
#define 	STEP_SIZE_SUB		0.03						// Bereite der Zwischenmarkierung
#define 	STEP_SIZE_XY		120							// Minimale Breite einer X-Beschriftung
#define 	STEP_SIZE_Z			120							// Minimale Breite einer Z-Beschriftung
#define 	STEP_SIZE_ZP		40							// Minimale Breite einer Z-Beschriftung der Box
#define 	STEP_OFFSET_X		5							// X-Abstand Skala Kurve
#define 	STEP_OFFSET_Z		5							// Y-Abstand Skala Kurve
#define 	STEP_OFFSET_FX		1							// X-Abstand Skala Beschriftung
#define 	STEP_OFFSET_FY		5							// Y-Abstand Skala Beschriftung
#define 	COLORBOX_WIDTH		8							// Breite der ColorBox

#define 	COLOR_BACKPLANE		BGR(255,255,255)			// Farbe fr den Hintergrund
#define 	COLOR_SHADOW		BGR(200,200,200)			// Farbe fr den Shatten
#define 	COLOR_GRID			BGR(64,64,64)				// Farbe fr den Ebenengitter
#define 	COLOR_SCALA			BGR(0,0,0)					// Farbe fr die Skalen-Striche
#define 	COLOR_SCALA_TEXT_X	BGR(120,70,0)				// Farbe fr die Skalen-Texte
#define 	COLOR_SCALA_TEXT_Y	BGR(100,0,100)				// Farbe fr die Skalen-Texte
#define 	COLOR_SCALA_TEXT_Z	BGR(200,0,0)				// Farbe fr die Skalen-Texte

#define		IS_NAN(a)		((((unsigned*)(&(a)))[1]&0x7FFF0000)>=0x7FF00000)

static		PlaneParams		sDefParam =
				{
				0.0,										// Von  -1  bis  +1 (Auf Ab)
				0.0,										// Von  -1  bis  +1 (Zoomen)
				0.0,										// Von -45 bis +45 Grad (Bild drehen)
				0.0,										// Von -45 bis +45 Grad
				36.0,										// Von -45 bis +45 Grad
				0.0,										// Position der Schattenebene
				2,											// Ebene der Schattenebene (0=X 1=Y 2=Z)
				};

static unsigned char aFont1[]=
	{
	#include "Image/Fonts/Arial8.h"
	};

static unsigned char aFont2[]=
	{
	#include "Image/Fonts/Arial16.h"
	};

static unsigned char aFont3[]=
	{
	#include "Image/Fonts/Arial36.h"
	};


//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
PlaneData::PlaneData()
{
	memset(this,0,sizeof(PlaneData));

}

//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
PlaneData::~PlaneData()
{
	if(pBufferZ)
		{
		delete pBufferZ;
		}

	pBufferZ  = 0;
	uBufSizeZ = 0;

}

//*****************************************************************************
//*
//*		PlaneCalcLinear
//*
//*****************************************************************************
//	Berechnet die linearen Schritte fr eine Skalierung
//	pScala			: Zeiger auf die Scala Daten
//	iSteps			: Unterteilung der gesammten Skala
//	Ergibt die zustzlichen Kommastellen
static int PlaneCalcLinear(PlaneScala *pScala,int iSteps)
{
double		dComma;
double		dSteps;
double		dValue;
double		dDelta;
int			iAdd;



	   dDelta = pScala->dMax-pScala->dMin;
	if(dDelta<=0.0)											// Prfe den Anzeigebereich
		{
		dDelta=1.0;
		if(pScala->dMin>1.0e06)pScala->dMin=0;
		pScala->dMax = pScala->dMin+dDelta;
		}

	dComma	= log10(dDelta/(double)iSteps);
	dValue	= floor(dComma);
	dComma -= dValue+0.02;
	dValue  = pow(10.0,dValue);

	if(dComma<=0.000000000)									// 1.0 Schritte
		{
		dSteps = dValue;
		iAdd   = 0;
		}
	else if(dComma<0.301029995)								// 2.0 Schritte
		{
		dSteps = dValue*2.0;
		iAdd   = 0;
		}
	else if(dComma<0.397940008)								// 2.5 Schritte
		{
		dSteps = dValue*2.5;
		iAdd   = 1;
		}
	else{													// 5.0 Schritte
		dSteps = dValue*5.0;
		iAdd   = 0;
		}


	dValue			  =  ceil(pScala->dMin/dSteps-0.00001);
	pScala->iIsLog	  =  0;
	pScala->iStepAt	  =  0;
	pScala->iStepCnt  =  1;
	pScala->aSteps[0] =  dSteps;
	pScala->dStart	  =  dValue*dSteps;
	pScala->dFact	  =  2.0/dDelta;
	pScala->dCalc	  =  1.0/pScala->dFact;
	pScala->dOffset	  = -1.0/pScala->dFact-pScala->dMin;



return iAdd;
}

//*****************************************************************************
//*
//*		PlaneCalcLog
//*
//*****************************************************************************
//	Berechnet die logarithmischen Schritte fr eine Skalierung
//	pScala			: Zeiger auf die Scala Daten
//	iSteps			: Breite der Skala in Ounkten
//	Ergibt die zustzlichen Kommastellen bzw. -3 bei linear Umschaltung sonst
//						-1 = keine Kommastelle
//						-2 = eine  Kommastelle
//						-3 = linear Umschaltung
static int PlaneCalcLog(PlaneScala *pScala,int iSteps)
{
double		dComma;
double		dSteps;
double		dValue;
double		dDelta;
int			iAdd;
int			iPos;


	   dValue = pScala->dMin;
	if(dValue<=1.0e-15)return -3;
	   dComma = pScala->dMax;
	if(dComma<=1.0e-15)return -3;
	   dDelta = log10(dComma/dValue);
	   dDelta/= (double)iSteps;
	if(dDelta<=0.15)return -3;

	dComma	= dDelta;
	dValue	= floor(dComma);
	dComma -= dValue+0.02;
	dSteps  = pow(10.0,dValue);

	if(dComma<0.17 && dValue>=1.0)
		{													// 10,100 Schritte
		iAdd				= -1;
		pScala->iStepCnt	=  1;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	= 10.0;
		}
	else if(dComma<0.22)									// 10,16,25,40,60,100 Schritte
		{
		iAdd				= -2;
		pScala->iStepCnt	=  5;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	=  1.6;
		pScala->aSteps[2]	=  2.5;
		pScala->aSteps[3]	=  4.0;
		pScala->aSteps[4]	=  6.0;
		pScala->aSteps[5]	= 10.0;
		}
	else if(dComma<0.25)									// 10,18,30,56,100 Schritte
		{
		iAdd				= -2;
		pScala->iStepCnt	=  4;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	=  1.8;
		pScala->aSteps[2]	=  3.0;
		pScala->aSteps[3]	=  5.6;
		pScala->aSteps[4]	= 10.0;
		}
	else if(dComma<0.33)									// 10,20,50,100 Schritte
		{
		iAdd				= -1;
		pScala->iStepCnt	=  3;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	=  2.0;
		pScala->aSteps[2]	=  5.0;
		pScala->aSteps[3]	= 10.0;
		}
	else if(dComma<0.49)									// 10,30,100 Schritte
		{
		iAdd				= -1;
		pScala->iStepCnt	=  2;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	=  3.0;
		pScala->aSteps[2]	= 10.0;
		}
	else{													// 10,100 Schritte
		iAdd				= -1;
		pScala->iStepCnt	=  1;
		pScala->aSteps[0]	=  1.0;
		pScala->aSteps[1]	= 10.0;
		}

	for(iPos=0;iPos<=pScala->iStepCnt;iPos++)				// Faktoren anfgen
		{
		pScala->aSteps[iPos] *= pow(dSteps,iPos);
		}

	dValue	= pow(10.0,floor(log10(pScala->dMin)));

	for(iPos=0;iPos<pScala->iStepCnt;iPos++)				// Suche Startposition
		{
		if(dValue*pScala->aSteps[iPos]*1.00001>=pScala->dMin)break;
		}

	if(iPos>=pScala->iStepCnt)
		{
		dValue *= pScala->aSteps[iPos];
		iPos    = 0;
		}

	pScala->dStart	=  dValue*pScala->aSteps[iPos];
	pScala->iStepAt =  iPos;
	pScala->iIsLog  =  1;
	pScala->dFact   =  2.0/log(pScala->dMax/pScala->dMin);
	pScala->dCalc	=  1.0/pScala->dFact;
	pScala->dOffset = -log(pScala->dMin)*pScala->dFact-1.0;


return iAdd;
}

//*****************************************************************************
//*
//*		ScalaText
//*
//*****************************************************************************
//	Erzeugt einen Text fr einen Skalenwert
static const char *ScalaText(PlaneScala *pData,char *pBuffer,double dValue)
{
double		dBase;
double		dCalc;
int			iBase;
int			iLen;
int			iMax;
int			iNum;


	   iBase = pData->iBase;
	if(iBase>=0)
		{
		if(pData->iComma)
			{
			iBase -= (pData->iComma+2)/3;
			}

		iMax  = pData->iPlaces;
		iMax -= pData->iComma;
		if(pData->iComma)iMax--;
		if(iMax<2)iMax=2;

		dBase = pow(1000.0,(double)(iBase-8));
		dCalc = dValue/dBase;

		if(dCalc<0.01 && dCalc>-0.01)						// Bei Null eine Verkrzung verhintern
			{
			   iNum = (int)log10(pData->aSteps[0]/dBase);
			if(iNum>=iMax)iBase++;
			}

		dCalc+= (dCalc<0.0)? -0.1:0.1;
		iLen  = vtostr((__int64)dCalc,pBuffer,VTS_SIGNED|VTS_ADDSUFFIX|VTS_CONSTSIZE|(iMax<<8)|(pData->iComma<<4)|(iBase<<12));

		if(iLen>0 && pBuffer[iLen-1]==' ')pBuffer[iLen-1]=0;

		return pBuffer+strwlen(pBuffer,0xE0);
		}


	iBase = 8;

	while(fabs(dValue)<9999.999)
		{
		dBase  = floor(dValue);
		dBase -= dValue;

		if(dBase ==0)break;
		if(dValue==0)break;
		if(iBase ==0)break;

		dValue *= 1000.0;
		iBase  -= 1;
		}

	while(fabs(dValue)>9999999.0)
		{
		if(dValue== 0)break;
		if(iBase ==15)break;

		dValue /= 1000.0;
		iBase  += 1;
		}

	if(pData->iComma && dValue<9999.0 && dValue>=1000.0 && fmod(dValue,100.0)==0)
		{
		pData->iComma=1;
		iMax = 2;
		}
	else{
		iMax = 5;
		}


	dCalc = (dValue<0.0)? dValue-0.1:dValue+0.1;
	iLen  = vtostr((__int64)dCalc,pBuffer,VTS_SIGNED|VTS_ADDSUFFIX|VTS_SHRINKCOMMA|(iMax<<8)|(pData->iComma<<4)|(iBase<<12));

	if(iLen>0 && pBuffer[iLen-1]==' ')pBuffer[iLen-1]=0;



return pBuffer;
}

//*****************************************************************************
//*
//*		SetText
//*
//*****************************************************************************
//	Zeichnet Punkte fr Texte
static void FC SetText(BitmapImage *pImage,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
double		dValZ;
int			iDifX;
int			iDifY;
int			iIdx;


	if(uPosX>=(unsigned)pData->iSizeX)return;
	if(uPosY>=(unsigned)pData->iSizeY)return;

	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx	= pData->iSizeX*uPosY+uPosX;

	if(dValZ>=pData->pBufferZ[iIdx])return;					// Mit Z-Buffer prfen

	pData->pBufferZ[iIdx]	= dValZ;

	BitmapImage::SetTextPixelSmoth(pImage,uPosX,uPosY,uColor);


}

//*****************************************************************************
//*
//*		SetGrid
//*
//*****************************************************************************
//	Zeichnet Punkte fr das Gitter
static void FC SetGrid(BitmapImage *pImage,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
double		dValZ;
int			iDifX;
int			iDifY;
int			iIdx;


	if(uPosX>=(unsigned)pData->iSizeX)return;
	if(uPosY>=(unsigned)pData->iSizeY)return;

	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx	= pData->iSizeX*uPosY+uPosX;

	if(dValZ>=pData->pBufferZ[iIdx])return;					// Mit Z-Buffer prfen

	pData->pBufferZ[iIdx]	= dValZ;

	pImage->SetPixel(uPosX,uPosY,uColor);


}

//*****************************************************************************
//*
//*		SetDual
//*
//*****************************************************************************
//	Zeichnest 2x2 Punkte fr die Skalen
static void FC SetDual(BitmapImage *pImage,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
unsigned	uColOld;
const int	aHlf[8*2] =
				{
				 -2, 0,
                 -2,-1,
				  1, 0,
				  1,-1,
				  0,-2,
				  1,-2,
				  0, 1,
				  1, 1,
				};
const int	aOff[4*2] =
				{
				 -1, 0,
				 -1,-1,
				  0, 0,
				  0,-1,
				};
double		dValZ;
unsigned	uOffX;
unsigned	uOffY;
int			iDifX;
int			iDifY;
int			iIdx;
int			iPos;
int			iNum;


	if(uPosX>=(unsigned)pData->iSizeX)return;
	if(uPosY>=(unsigned)pData->iSizeY)return;

	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx = pData->iSizeX*uPosY+uPosX;


	for(iPos=0;iPos<4;iPos++)
		{
		uOffX	= uPosX+aOff[iPos*2+0];
		uOffY	= uPosY+aOff[iPos*2+1];

		if(uOffX>=(unsigned)pData->iSizeX)continue;
		if(uOffY>=(unsigned)pData->iSizeY)continue;

		iNum	= iIdx;
		iNum   += aOff[iPos*2+0];
		iNum   += aOff[iPos*2+1]*pData->iSizeX;

		if(dValZ>=pData->pBufferZ[iNum])continue;			// Mit Z-Buffer prfen

		pData->pBufferZ[iNum]	= dValZ;
		pImage->SetPixel(uOffX,uOffY,uColor);
		}


	for(iPos=0;iPos<8;iPos++)								// Durchsichtige Pixel
		{
		uOffX	= uPosX+aHlf[iPos*2+0];
		uOffY	= uPosY+aHlf[iPos*2+1];

		if(uOffX>=(unsigned)pData->iSizeX)continue;
		if(uOffY>=(unsigned)pData->iSizeY)continue;

		iNum	= iIdx;
		iNum   += aHlf[iPos*2+0];
		iNum   += aHlf[iPos*2+1]*pData->iSizeX;

		if(dValZ>=pData->pBufferZ[iNum])continue;			// Mit Z-Buffer prfen

		uColOld = pImage->GetPixel(uOffX,uOffY);
		uColOld =(uColOld&0x00FEFEFE)>>1;
		uColOld+=(uColor &0x00FEFEFE)>>1;;

		pImage->SetPixel(uOffX,uOffY,uColOld);
		}


}

//*****************************************************************************
//*
//*		SetQuard
//*
//*****************************************************************************
//	Zeichnest 4x4 Punkte fr die Skalen
static void FC SetQuard(BitmapImage *pImage,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
unsigned	uColOld;
const int	aHlf[12*2] =
				{
				 -3, 0,
                 -3,-1,
				  2, 0,
				  2,-1,
				  0,-3,
				  1,-3,
				  0, 2,
				  1, 2,
				 -2, 1,
				 -2,-2,
				  1, 1,
				  1,-2,
				};
const int	aOff[12*2] =
				{
				 -2, 0,
                 -2,-1,
				 -1, 0,
				 -1,-1,
				  0, 0,
				  0,-1,
				  1, 0,
				  1,-1,
				  0,-2,
				 -1,-2,
				  0, 1,
				 -1, 1,
				};
double		dValZ;
unsigned	uOffX;
unsigned	uOffY;
int			iDifX;
int			iDifY;
int			iIdx;
int			iPos;
int			iNum;


	if(uPosX>=(unsigned)pData->iSizeX)return;
	if(uPosY>=(unsigned)pData->iSizeY)return;

	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx = pData->iSizeX*uPosY+uPosX;


	for(iPos=0;iPos<12;iPos++)
		{
		uOffX	= uPosX+aOff[iPos*2+0];
		uOffY	= uPosY+aOff[iPos*2+1];

		if(uOffX>=(unsigned)pData->iSizeX)continue;
		if(uOffY>=(unsigned)pData->iSizeY)continue;

		iNum	= iIdx;
		iNum   += aOff[iPos*2+0];
		iNum   += aOff[iPos*2+1]*pData->iSizeX;

		if(dValZ>=pData->pBufferZ[iNum])continue;			// Mit Z-Buffer prfen

		pData->pBufferZ[iNum]	= dValZ;
		pImage->SetPixel(uOffX,uOffY,uColor);
		}

	for(iPos=0;iPos<12;iPos++)								// Durchsichtige Pixel
		{
		uOffX	= uPosX+aHlf[iPos*2+0];
		uOffY	= uPosY+aHlf[iPos*2+1];

		if(uOffX>=(unsigned)pData->iSizeX)continue;
		if(uOffY>=(unsigned)pData->iSizeY)continue;

		iNum	= iIdx;
		iNum   += aHlf[iPos*2+0];
		iNum   += aHlf[iPos*2+1]*pData->iSizeX;

		if(dValZ>=pData->pBufferZ[iNum])continue;			// Mit Z-Buffer prfen

		uColOld = pImage->GetPixel(uOffX,uOffY);
		uColOld =(uColOld&0x00FEFEFE)>>1;
		uColOld+=(uColor &0x00FEFEFE)>>1;;

		pImage->SetPixel(uOffX,uOffY,uColOld);
		}


}

//*****************************************************************************
//*
//*		SetTriangel
//*
//*****************************************************************************
//	Zeichnet Punkte fr Dreiecke
static void FC SetTriangel(BitmapImage *pImage,unsigned char *pPtr,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
D3Vector3	sVectorA;
D3Vector3	sVectorB;
double		dValZ;
unsigned	uGreen;
unsigned	uBlue;
unsigned	uRed;
int			iDifX;
int			iDifY;
int			iIdx;



	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx	= pData->iSizeX*uPosY+uPosX;

	if(dValZ>=pData->pBufferZ[iIdx])return;					// Mit Z-Buffer prfen

	pData->pBufferZ[iIdx]	= dValZ;

	sVectorA.sVector.x	= (int)uPosX;
	sVectorA.sVector.y	= (int)uPosY;
	sVectorA.sVector.z	=      dValZ;

	VectorTransformCoord(&sVectorB,&pData->sDiv,&sVectorA);


	if(sVectorB.sVector.z<-1.0)
		{
		if(sVectorB.sVector.z>-1.25)
				uRed = (int)((sVectorB.sVector.z+1.0)*-1020.0);
		else	uRed = 255;

		uBlue	= 255-uRed/2;

		uColor  = (((255-uRed  )*pData->uWhite)/WHITE_DIF+uRed  )<<16;
		uColor |= (((255       )*pData->uWhite)/WHITE_DIF       )<<8;
		uColor |= (((255-uBlue )*pData->uWhite)/WHITE_DIF+uBlue );
		}
	else if(sVectorB.sVector.z<=1.0)
		{
		uBlue	= (int)((1.0-sVectorB.sVector.z)*127.5);
		uGreen	= 255-uBlue;

		uColor  = (((255       )*pData->uWhite)/WHITE_DIF       )<<16;
		uColor |= (((255-uGreen)*pData->uWhite)/WHITE_DIF+uGreen)<<8;
		uColor |= (((255-uBlue )*pData->uWhite)/WHITE_DIF+uBlue );
		}
	else{
		if(sVectorB.sVector.z<1.25)
				uRed = (int)((sVectorB.sVector.z-1.0)*1020.0);
		else	uRed = 255;

		uGreen = 255-uRed;

		uColor  = (((255-uRed  )*pData->uWhite)/WHITE_DIF+uRed  )<<16;
		uColor |= (((255-uGreen)*pData->uWhite)/WHITE_DIF+uGreen)<<8;
		uColor |= (((255       )*pData->uWhite)/WHITE_DIF       );
		}


	pImage->SetPixel(pPtr,uPosX,uColor);


}

//*****************************************************************************
//*
//*		PutText
//*
//*****************************************************************************
//	Zeichet eine Text auf die 3D-Ebene (die 3D-Aurichtung muss in pData gesetzt sein)
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image
//	pText		: Zeiger auf den Text
//	pBase		: Ausgangspunkt des Textes
//	pSize		: Gre der einzelnen Punkte
//	uColor		: Textfarbe
//	iPlane		: Lage (0=X 1=Y 2=Z 3=Z mittig)
//	Ergibt die Anzahl der zu berechnenten Punkte
static void PutText(PlaneData *pData,BitmapImage *pImage,const char *pText,D3Vector3 *pBase,double *pSize,unsigned uColor,int iPlane)
{
D3Vector3		sVector;
D3Vector3		sCoord;
BitmapImage		sTemp;
int				iPosX;
int				iPosY;
int				iMax;


	sTemp.SetSize(8,512,56);
	sTemp.SetFont(aFont3);
	sTemp.SetPixelFunction(TEXT_NONE  ,0xFF,1);
	sTemp.SetPixelFunction(TEXT_NORMAL,0xFF,0);
	iMax = sTemp.Print(0,0,pText);

	switch(iPlane)
		{
	case 0:	for(iPosY=0;iPosY<56  ;iPosY++)
			for(iPosX=0;iPosX<iMax;iPosX++)
				{
				if(!sTemp.GetPixel(iPosX,iPosY))continue;

				sVector.sData[0]	= pBase->sData[0]+pSize[0]*(double)(24-iPosY);
				sVector.sData[1]	= pBase->sData[1]+pSize[1]*(double)(iPosX);
				sVector.sData[2]	= pBase->sData[2]+pSize[2]*(double)(iPosX);

				VectorTransformCoord(&sCoord,&pData->sMul,&sVector);
				SetText(pImage,(int)sCoord.sVector.x,(int)sCoord.sVector.y,uColor);
				}

			break;

	case 1:	for(iPosY=0;iPosY<56  ;iPosY++)
			for(iPosX=0;iPosX<iMax;iPosX++)
				{
				if(!sTemp.GetPixel(iPosX,iPosY))continue;

				sVector.sData[0]	= pBase->sData[0]+pSize[0]*(double)(iMax-iPosX);
				sVector.sData[1]	= pBase->sData[1]+pSize[1]*(double)(24-iPosY);
				sVector.sData[2]	= pBase->sData[2]+pSize[2]*(double)(iPosX);

				VectorTransformCoord(&sCoord,&pData->sMul,&sVector);
				SetText(pImage,(int)sCoord.sVector.x,(int)sCoord.sVector.y,uColor);
				}

			break;

	case 2:	for(iPosY=0;iPosY<56  ;iPosY++)
			for(iPosX=0;iPosX<iMax;iPosX++)
				{
				if(!sTemp.GetPixel(iPosX,iPosY))continue;

				sVector.sData[0]	= pBase->sData[0]+pSize[0]*(double)(iPosX);
				sVector.sData[1]	= pBase->sData[1]+pSize[1]*(double)(iPosX);
				sVector.sData[2]	= pBase->sData[2]+pSize[2]*(double)(iPosY-24);

				VectorTransformCoord(&sCoord,&pData->sMul,&sVector);
				SetText(pImage,(int)sCoord.sVector.x,(int)sCoord.sVector.y,uColor);
				}

			break;

	case 3:	for(iPosY=0;iPosY<56  ;iPosY++)
			for(iPosX=0;iPosX<iMax;iPosX++)
				{
				if(!sTemp.GetPixel(iPosX,iPosY))continue;

				sVector.sData[0]	= pBase->sData[0]+pSize[0]*(double)(iPosX-iMax/2);
				sVector.sData[1]	= pBase->sData[1]+pSize[1]*(double)(iPosX-iMax/2);
				sVector.sData[2]	= pBase->sData[2]+pSize[2]*(double)(iPosY-24);

				VectorTransformCoord(&sCoord,&pData->sMul,&sVector);
				SetText(pImage,(int)sCoord.sVector.x,(int)sCoord.sVector.y,uColor);
				}

			break;
		}

}

//*****************************************************************************
//*
//*		SetShadow
//*
//*****************************************************************************
//	Zeichnet durchsichtige Pixel fr die Schattenebene
static void FC SetShadow(BitmapImage *pImage,unsigned char *pPtr,unsigned uPosX,unsigned uPosY,unsigned uColor)
{
PlaneData  *pData = (PlaneData*)pImage->GetUserPtr();
D3Vector3	sVectorA;
D3Vector3	sVectorB;
unsigned	uColNow;
double		dValZ;
int			iDifX;
int			iDifY;
int			iIdx;


	iDifX	= pData->iPosX-uPosX;
	iDifY	= pData->iPosY-uPosY;
	dValZ	= pData->dAddZ+pData->dFactX*(double)iDifX+pData->dFactY*(double)iDifY;

	if(dValZ<=0.0)return;									// Zu nah am Auge

	iIdx	= pData->iSizeX*uPosY+uPosX;

	if(dValZ>=pData->pBufferZ[iIdx])return;					// Mit Z-Buffer prfen

	pData->pBufferZ[iIdx] = dValZ;

	sVectorA.sVector.x	= (int)uPosX;
	sVectorA.sVector.y	= (int)uPosY;
	sVectorA.sVector.z	=      dValZ;

	VectorTransformCoord(&sVectorB,&pData->sDiv,&sVectorA);


	uColNow	  =	pImage->GetPixel(pPtr,uPosX);
	uColNow  &= 0x00FCFCFC;
	uColNow >>= 2;
	uColNow  *= 3;
	uColor   &= 0x00FCFCFC;
	uColor  >>= 2;
	uColor   += uColNow;

	pImage->SetPixel(pPtr,uPosX,uColor);




}


//*****************************************************************************
//*
//*		PlaneDrawGrid
//*
//*****************************************************************************
//	Zeichnet eine dreidimensionale Ebene
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	dOffset		: Ist die Verschub zum Auge
void PlaneDrawGrid(PlaneData *pData,BitmapImage *pImage,double dOffset)
{
D3Vector3		sVectorA;
D3Vector3		sVectorB;
D3Vector3		sVector;
D3Vector3	   *pCoord;
int				iPosX;
int				iPosY;
int				iPos;



	pCoord = pData->pGridVects;


	for(iPosY=1;iPosY<pData->iSteps;iPosY++)				// Gitter zeichnen
	for(iPosX=1;iPosX<pData->iSteps;iPosX++)
		{
		sVectorA.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.x;
		sVectorA.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		sVectorA.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.y;
		sVectorA.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;
		sVectorA.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.z;
		sVectorA.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.z;

		sVectorB.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.x;
		sVectorB.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		sVectorB.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.y;
		sVectorB.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;
		sVectorB.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.z;
		sVectorB.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.z;

		VectorCross(&sVector,&sVectorA,&sVectorB);

		pData->dAddZ	= pCoord[iPosY*GRID_MAX+iPosX].sVector.z+dOffset;
		pData->dFactX	= sVector.sVector.x;
		pData->dFactX  /= sVector.sVector.z;
		pData->dFactY	= sVector.sVector.y;
		pData->dFactY  /= sVector.sVector.z;
		pData->iPosX	= (int)pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		pData->iPosY	= (int)pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;

		if(!(iPosY&pData->uGridMask))
		pImage->Line((int)pCoord[iPosY*GRID_MAX+iPosX-1].sVector.x,(int)pCoord[iPosY*GRID_MAX+iPosX-1].sVector.y,
					 (int)pCoord[iPosY*GRID_MAX+iPosX  ].sVector.x,(int)pCoord[iPosY*GRID_MAX+iPosX  ].sVector.y,COLOR_GRID,SetGrid);

		if(!(iPosX&pData->uGridMask))
		pImage->Line((int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX].sVector.x,(int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX].sVector.y,
					 (int)pCoord[iPosY*GRID_MAX+iPosX         ].sVector.x,(int)pCoord[iPosY*GRID_MAX+iPosX         ].sVector.y,COLOR_GRID,SetGrid);
		}


	for(iPos=1;iPos<pData->iSteps;iPos++)					// Gitter zeichnen erste Reihen
		{
		sVectorA.sVector.x  = pCoord[iPos-1       ].sVector.x;
		sVectorA.sVector.x -= pCoord[iPos         ].sVector.x;
		sVectorA.sVector.y  = pCoord[iPos-1       ].sVector.y;
		sVectorA.sVector.y -= pCoord[iPos         ].sVector.y;
		sVectorA.sVector.z  = pCoord[iPos-1       ].sVector.z;
		sVectorA.sVector.z -= pCoord[iPos         ].sVector.z;

		sVectorB.sVector.x  = pCoord[iPos+GRID_MAX].sVector.x;
		sVectorB.sVector.x -= pCoord[iPos         ].sVector.x;
		sVectorB.sVector.y  = pCoord[iPos+GRID_MAX].sVector.y;
		sVectorB.sVector.y -= pCoord[iPos         ].sVector.y;
		sVectorB.sVector.z  = pCoord[iPos+GRID_MAX].sVector.z;
		sVectorB.sVector.z -= pCoord[iPos         ].sVector.z;

		VectorCross(&sVector,&sVectorA,&sVectorB);

		pData->dAddZ	= pCoord[iPos].sVector.z+dOffset;
		pData->dFactX	= sVector.sVector.x;
		pData->dFactX  /= sVector.sVector.z;
		pData->dFactY	= sVector.sVector.y;
		pData->dFactY  /= sVector.sVector.z;
		pData->iPosX	= (int)pCoord[iPos].sVector.x;
		pData->iPosY	= (int)pCoord[iPos].sVector.y;

		pImage->Line((int)pCoord[iPos-1].sVector.x,(int)pCoord[iPos-1].sVector.y,
					 (int)pCoord[iPos  ].sVector.x,(int)pCoord[iPos  ].sVector.y,COLOR_GRID,SetGrid);


		sVectorA.sVector.x  = pCoord[iPos*GRID_MAX-GRID_MAX].sVector.x;
		sVectorA.sVector.x -= pCoord[iPos*GRID_MAX         ].sVector.x;
		sVectorA.sVector.y  = pCoord[iPos*GRID_MAX-GRID_MAX].sVector.y;
		sVectorA.sVector.y -= pCoord[iPos*GRID_MAX         ].sVector.y;
		sVectorA.sVector.z  = pCoord[iPos*GRID_MAX-GRID_MAX].sVector.z;
		sVectorA.sVector.z -= pCoord[iPos*GRID_MAX         ].sVector.z;

		sVectorB.sVector.x  = pCoord[iPos*GRID_MAX+1       ].sVector.x;
		sVectorB.sVector.x -= pCoord[iPos*GRID_MAX         ].sVector.x;
		sVectorB.sVector.y  = pCoord[iPos*GRID_MAX+1       ].sVector.y;
		sVectorB.sVector.y -= pCoord[iPos*GRID_MAX         ].sVector.y;
		sVectorB.sVector.z  = pCoord[iPos*GRID_MAX+1       ].sVector.z;
		sVectorB.sVector.z -= pCoord[iPos*GRID_MAX         ].sVector.z;

		VectorCross(&sVector,&sVectorA,&sVectorB);

		pData->dAddZ	= pCoord[iPos*GRID_MAX].sVector.z;
		pData->dFactX	= sVector.sVector.x;
		pData->dFactX  /= sVector.sVector.z;
		pData->dFactY	= sVector.sVector.y;
		pData->dFactY  /= sVector.sVector.z;
		pData->iPosX	= (int)pCoord[iPos*GRID_MAX].sVector.x;
		pData->iPosY	= (int)pCoord[iPos*GRID_MAX].sVector.y;

		pImage->Line((int)pCoord[iPos*GRID_MAX-GRID_MAX].sVector.x,(int)pCoord[iPos*GRID_MAX-GRID_MAX].sVector.y,
					 (int)pCoord[iPos*GRID_MAX         ].sVector.x,(int)pCoord[iPos*GRID_MAX         ].sVector.y,COLOR_GRID,SetGrid);
		}

}

//*****************************************************************************
//*
//*		PlaneDrawShadow
//*
//*****************************************************************************
//	Zeichnet eine dreidimensionale Ebene
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	dOffset		: Ist die Position der Flche auf der Z-Achse
//	iNum		: Ist die Ebene (0=X 1=Y 2=Z)
void PlaneDrawShadow(PlaneData *pData,BitmapImage *pImage,double dOffset=0.0,int iNum=2,unsigned uColor=COLOR_SHADOW)
{
double		   *pBufferZ;
double		   *pBufferO;
D3Vector3		sVectorA;
D3Vector3		sVectorB;
D3Vector3		sVector;
D3Vector3		aVector[4];
D3Vector3		aCoord [4];
int				aArray[10];
double			dValue;
int				iPos;
int				iCnt;


	if(!pImage->IsRGB() || iNum<0 || iNum>2)return;

	   dValue	= PlaneValToPos(pData,dOffset,iNum);
	if(dValue<-1.01)return;
	if(dValue> 1.01)return;


	aVector[0].sData[(iNum+1)%3]	= -1.0;
	aVector[0].sData[(iNum+2)%3]	= -1.0;
	aVector[0].sData[(iNum  )%3]	=  dValue;
	aVector[1].sData[(iNum+1)%3]	=  1.0;
	aVector[1].sData[(iNum+2)%3]	= -1.0;
	aVector[1].sData[(iNum  )%3]	=  dValue;
	aVector[2].sData[(iNum+1)%3]	=  1.0;
	aVector[2].sData[(iNum+2)%3]	=  1.0;
	aVector[2].sData[(iNum  )%3]	=  dValue;
	aVector[3].sData[(iNum+1)%3]	= -1.0;
	aVector[3].sData[(iNum+2)%3]	=  1.0;
	aVector[3].sData[(iNum  )%3]	=  dValue;


	for(iPos=0;iPos<4;iPos++)
		{
		VectorTransformCoord(aCoord+iPos,&pData->sMul,aVector+iPos);
		}


	aArray[0]	= (int)aCoord[2+0].sVector.x;			// Dreiecke zeichnen
	aArray[1]	= (int)aCoord[2+0].sVector.y;
	aArray[2]	= (int)aCoord[2+1].sVector.x;
	aArray[3]	= (int)aCoord[2+1].sVector.y;
	aArray[4]	= (int)aCoord[0+0].sVector.x;
	aArray[5]	= (int)aCoord[0+0].sVector.y;
	aArray[6]	= (int)aCoord[0+1].sVector.x;
	aArray[7]	= (int)aCoord[0+1].sVector.y;
	aArray[8]	= aArray[0];
	aArray[9]	= aArray[1];

	sVectorA.sVector.x  = aCoord[2+0].sVector.x;
	sVectorA.sVector.x -= aCoord[2+1].sVector.x;
	sVectorA.sVector.y  = aCoord[2+0].sVector.y;
	sVectorA.sVector.y -= aCoord[2+1].sVector.y;
	sVectorA.sVector.z  = aCoord[2+0].sVector.z;
	sVectorA.sVector.z -= aCoord[2+1].sVector.z;

	sVectorB.sVector.x  = aCoord[0+1].sVector.x;
	sVectorB.sVector.x -= aCoord[2+1].sVector.x;
	sVectorB.sVector.y  = aCoord[0+1].sVector.y;
	sVectorB.sVector.y -= aCoord[2+1].sVector.y;
	sVectorB.sVector.z  = aCoord[0+1].sVector.z;
	sVectorB.sVector.z -= aCoord[2+1].sVector.z;

	VectorCross(&sVector ,&sVectorA,&sVectorB);

	pData->dAddZ	= aCoord[2+1].sVector.z;
	pData->dFactX	= sVector.sVector.x;
	pData->dFactX  /= sVector.sVector.z;
	pData->dFactY	= sVector.sVector.y;
	pData->dFactY  /= sVector.sVector.z;
	pData->iPosX	= aArray[2];
	pData->iPosY	= aArray[3];

	iCnt			= pData->iSizeX*pData->iSizeY;
	pBufferZ		= new double[iCnt];
	pBufferO		= pData->pBufferZ;

	memcpy(pBufferZ,pData->pBufferZ,iCnt*sizeof(double));

	pData->pBufferZ	= pBufferZ;

	pImage->Triangle(aArray,uColor,SetShadow);

	pData->dAddZ   += 0.000001;

	pImage->Triangle(aArray+4,uColor,SetShadow);

	pData->dAddZ   -= 0.000001;


	delete pBufferZ;
	pData->pBufferZ	= pBufferO;


}

//*****************************************************************************
//*
//*		PlaneDrawScala
//*
//*****************************************************************************
//	Zeichnet eine Skala
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	pScale		: Zeiger auf die Skala
//	uColor		: Textfarbe
void PlaneDrawScala(PlaneData *pData,BitmapImage *pImage,PlaneScala *pScale,unsigned uColor)
{
char			cText[64];
int				aArray[10];
PixelFunction	pPixel;
D3Vector3		sVectorA;
D3Vector3		sVectorB;
D3Vector3		sVector;
D3Vector3		aVector[3];
D3Vector3		aCoord [3];
double			aFontSz[3];
double			dBase1;
double			dBase2;
double			dSlot1;
double			dSlot2;
double			dVal;
double			dSub;
double			dPos;
double			dBit;
double			dMax;
double			dMin;
double			dLen;
int				iSlot1;
int				iSlot2;
int				iLen;
int				iIdx;
int				iNum;
int				iMul;

								  {iMul=1;pPixel = SetGrid ;}
	if(pData->uOptions&PDR_DOUBLE){iMul=2;pPixel = SetDual ;}
	if(pData->uOptions&PDR_QUARD ){iMul=4;pPixel = SetQuard;}


	switch(pScale->iMode)
		{
	case 0:													// X-Scala

		iSlot1					=  1;
		iSlot2					=  2;
		dSlot1					= -1.0;
		dSlot2					=  0.0;
		dBase1					= -1.0;
		dBase2					= -1.0;
		aVector[0].sVector.x	= -1.0;
		aVector[0].sVector.y	= -1.0;
		aVector[0].sVector.z	= -1.0;
		aVector[1].sVector.x	=  1.0;
		aVector[1].sVector.y	= -1.0;
		aVector[1].sVector.z	= -1.0;
		aVector[2].sVector.x	=  1.0;
		aVector[2].sVector.y	= -1.0-1.0;
		aVector[2].sVector.z	= -1.0;
		aFontSz[iSlot1]			=  dSlot1;
		aFontSz[iSlot2]			=  dSlot2;
		aFontSz[pScale->iMode]	=  1.0;
		break;

	case 1:													// Z-Scala

		iSlot1					=  0;
		iSlot2					=  2;
		dSlot1					= -1.0;
		dSlot2					=  0.0;
		dBase1					= -1.0;
		dBase2					= -1.0;
		aVector[0].sVector.x	= -1.0;
		aVector[0].sVector.y	= -1.0;
		aVector[0].sVector.z	= -1.0;
		aVector[1].sVector.x	= -1.0;
		aVector[1].sVector.y	=  1.0;
		aVector[1].sVector.z	= -1.0;
		aVector[2].sVector.x	= -1.0-1.0;
		aVector[2].sVector.y	=  1.0;
		aVector[2].sVector.z	= -1.0;
		aFontSz[iSlot1]			=  dSlot1;
		aFontSz[iSlot2]			=  dSlot2;
		aFontSz[pScale->iMode]	=  1.0;

		break;

	case 2:													// Z-Scala

		iSlot1					=  0;
		iSlot2					=  1;
		dSlot1					= -sin(pData->dRollZ-D3_PI*0.5);
		dSlot2					= -cos(pData->dRollZ-D3_PI*0.5);
		dBase1					= -1.0;
		dBase2					= -1.0;
		aVector[0].sVector.x	= -1.0;
		aVector[0].sVector.y	= -1.0;
		aVector[0].sVector.z	= -1.0;
		aVector[1].sVector.x	= -1.0;
		aVector[1].sVector.y	= -1.0;
		aVector[1].sVector.z	=  1.0;
		aVector[2].sVector.x	= -1.0-dSlot1;
		aVector[2].sVector.y	= -1.0-dSlot2;
		aVector[2].sVector.z	=  1.0;
		aFontSz[iSlot1]			=  dSlot1;
		aFontSz[iSlot2]			=  dSlot2;
		aFontSz[pScale->iMode]	= -1.0;

		break;

	default: return;
		}


	aFontSz[0] *= pData->dFontScale*(double)iMul;
	aFontSz[1] *= pData->dFontScale*(double)iMul;
	aFontSz[2] *= pData->dFontScale*(double)iMul;


	VectorTransformCoord(aCoord+0,&pData->sMul,aVector+0);
	VectorTransformCoord(aCoord+1,&pData->sMul,aVector+1);
	VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

	aArray[0]	= (int)aCoord[0].sVector.x;					// Dreiecke zeichnen
	aArray[1]	= (int)aCoord[0].sVector.y;
	aArray[2]	= (int)aCoord[1].sVector.x;
	aArray[3]	= (int)aCoord[1].sVector.y;
	aArray[4]	= (int)aCoord[2].sVector.x;
	aArray[5]	= (int)aCoord[2].sVector.y;

	sVectorA.sVector.x  = aCoord[0].sVector.x;
	sVectorA.sVector.x -= aCoord[1].sVector.x;
	sVectorA.sVector.y  = aCoord[0].sVector.y;
	sVectorA.sVector.y -= aCoord[1].sVector.y;
	sVectorA.sVector.z  = aCoord[0].sVector.z;
	sVectorA.sVector.z -= aCoord[1].sVector.z;

	sVectorB.sVector.x  = aCoord[2].sVector.x;
	sVectorB.sVector.x -= aCoord[1].sVector.x;
	sVectorB.sVector.y  = aCoord[2].sVector.y;
	sVectorB.sVector.y -= aCoord[2].sVector.y;
	sVectorB.sVector.z  = aCoord[1].sVector.z;
	sVectorB.sVector.z -= aCoord[2].sVector.z;

	VectorCross(&sVector ,&sVectorA,&sVectorB);

	pData->dAddZ	= aCoord[2].sVector.z;
	pData->dFactX	= sVector.sVector.x;
	pData->dFactX  /= sVector.sVector.z;
	pData->dFactY	= sVector.sVector.y;
	pData->dFactY  /= sVector.sVector.z;
	pData->iPosX	= aArray[2];
	pData->iPosY	= aArray[3];

															// Hauptlinie
	pImage->Line(aArray[0],aArray[1],aArray[2],aArray[3],COLOR_SCALA,pPixel);


	if(pData->uOptions&(PDR_TEXT_X<<pScale->iMode))			// Einheien Beschriftung
		{
		iLen = strlcpy(cText,pScale->cVarName,sizeof(cText));
		if(pScale->cUnitName[0])
			{
			iLen += sprintf(cText+iLen," [%s]",pScale->cUnitName);
			}

		dVal						    = STEP_SIZE_NULL*pData->dFontScale*700.0*(double)iMul;
		aVector[2].sData[pScale->iMode]	= aVector[1].sData[pScale->iMode]+dVal;
		aVector[2].sData[iSlot1]		= dBase1;
		aVector[2].sData[iSlot2]		= dBase2;

		VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);
		PutText(pData,pImage,cText,aVector+2,aFontSz,uColor,(pScale->iMode==2)? 3:pScale->iMode);
		}

	if(pScale->iIsLog)
		{
		dVal	= pScale->dStart;
		dMax	= pScale->dMax*1.00001;
		dMin	= pScale->dMin*0.99999;
		iIdx	= pScale->iStepAt;

		for(iNum=0;dVal<=dMax && iNum<64;iNum++)
			{
			dPos = PlaneValToPos(pData,dVal,pScale->iMode);

			if(dPos>=-1.000001 && dPos<=1.000001)			// Groe Teilungen
				{
				dLen = STEP_SIZE_MAIN;

				aVector[1].sData[pScale->iMode]	= dPos;
				aVector[2].sData[pScale->iMode]	= dPos;
				aVector[2].sData[iSlot1]		= dBase1+dSlot1*dLen;
				aVector[2].sData[iSlot2]		= dBase2+dSlot2*dLen;

				VectorTransformCoord(aCoord+1,&pData->sMul,aVector+1);
				VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

				aArray[2]	= (int)aCoord[1].sVector.x;
				aArray[3]	= (int)aCoord[1].sVector.y;
				aArray[4]	= (int)aCoord[2].sVector.x;
				aArray[5]	= (int)aCoord[2].sVector.y;

				pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);

				if(pScale->iDoubleMark)
					{
					aVector[2].sData[iSlot1]	= dBase1-dSlot1*dLen;
					aVector[2].sData[iSlot2]	= dBase2-dSlot2*dLen;

					VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

					aArray[4]	= (int)aCoord[2].sVector.x;
					aArray[5]	= (int)aCoord[2].sVector.y;

					pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);
					}

				if(pData->uOptions&(PDR_TEXT_X<<pScale->iMode))
					{
					dLen						= STEP_SIZE_NULL+0.005;
					aVector[2].sData[iSlot1]	= dBase1+dSlot1*dLen;
					aVector[2].sData[iSlot2]	= dBase2+dSlot2*dLen;
					ScalaText(pScale,cText,dVal);
					PutText(pData,pImage,cText,aVector+2,aFontSz,uColor,pScale->iMode);
					}
				}

			dVal /= pScale->aSteps[iIdx];iIdx++;
			dVal *=	pScale->aSteps[iIdx];

			if(iIdx>=pScale->iStepCnt)iIdx=0;
			}
		}
	else{													// Lineare Teilung
		dVal	= pScale->dStart;
		dSub	= pScale->dStart-pScale->aSteps[0]*0.5;
		dMax	= pScale->dMax  +pScale->aSteps[0]*0.6;
		dMin	= pScale->dMin;
		dBit	= pScale->dCalc*0.002;

		for(iNum=0;iNum<64;iNum++)
			{
			if(dVal>=dMax)break;

			dPos = PlaneValToPos(pData,dVal,pScale->iMode);

			if(dPos>=-1.000001 && dPos<=1.000001)			// Groe Teilungen
				{
				dLen = (dVal<dBit && dVal>=-dBit)? STEP_SIZE_NULL:STEP_SIZE_MAIN;

				aVector[1].sData[pScale->iMode]	= dPos;
				aVector[2].sData[pScale->iMode]	= dPos;
				aVector[2].sData[iSlot1]		= dBase1+dSlot1*dLen;
				aVector[2].sData[iSlot2]		= dBase2+dSlot2*dLen;

				VectorTransformCoord(aCoord+1,&pData->sMul,aVector+1);
				VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

				aArray[2]	= (int)aCoord[1].sVector.x;
				aArray[3]	= (int)aCoord[1].sVector.y;
				aArray[4]	= (int)aCoord[2].sVector.x;
				aArray[5]	= (int)aCoord[2].sVector.y;

				pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);

				if(pScale->iDoubleMark)
					{
					aVector[2].sData[iSlot1]	= dBase1-dSlot1*dLen;
					aVector[2].sData[iSlot2]	= dBase2-dSlot2*dLen;

					VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

					aArray[4]	= (int)aCoord[2].sVector.x;
					aArray[5]	= (int)aCoord[2].sVector.y;

					pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);
					}

				if(pData->uOptions&(PDR_TEXT_X<<pScale->iMode))
					{
					dLen						= STEP_SIZE_NULL+0.005;
					aVector[2].sData[iSlot1]	= dBase1+dSlot1*dLen;
					aVector[2].sData[iSlot2]	= dBase2+dSlot2*dLen;
					ScalaText(pScale,cText,dVal);
					PutText(pData,pImage,cText,aVector+2,aFontSz,uColor,pScale->iMode);
					}
				}

			dPos = PlaneValToPos(pData,dSub,pScale->iMode);

			if(dPos>=-1.000001 && dPos<=1.000001)			// Kleine Teilungen
				{
				dLen = (dPos<dBit && dPos>=-dBit)? STEP_SIZE_NULL:STEP_SIZE_SUB;

				aVector[1].sData[pScale->iMode]	= dPos;
				aVector[2].sData[pScale->iMode]	= dPos;
				aVector[2].sData[iSlot1]		= dBase1+dSlot1*dLen;
				aVector[2].sData[iSlot2]		= dBase2+dSlot2*dLen;

				VectorTransformCoord(aCoord+1,&pData->sMul,aVector+1);
				VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

				aArray[2]	= (int)aCoord[1].sVector.x;
				aArray[3]	= (int)aCoord[1].sVector.y;
				aArray[4]	= (int)aCoord[2].sVector.x;
				aArray[5]	= (int)aCoord[2].sVector.y;

				pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);

				if(pScale->iDoubleMark)
					{
					aVector[2].sData[iSlot1]	= dBase1-dSlot1*dLen;
					aVector[2].sData[iSlot2]	= dBase2-dSlot2*dLen;

					VectorTransformCoord(aCoord+2,&pData->sMul,aVector+2);

					aArray[4]	= (int)aCoord[2].sVector.x;
					aArray[5]	= (int)aCoord[2].sVector.y;

					pImage->Line(aArray[4],aArray[5],aArray[2],aArray[3],COLOR_SCALA,pPixel);
					}
				}


			dVal += pScale->aSteps[0];
			dSub += pScale->aSteps[0];
			}
		}


}

//*****************************************************************************
//*
//*		PlaneDrawColorField
//*
//*****************************************************************************
//	Zeichnet ein Farbenrechteck anzeigen
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	pValues		: Berechnete Z-Werte
static void PlaneDrawColorField(PlaneData *pData,BitmapImage *pImage,double *pValues)
{
BitmapImage		sField;
BitmapImage		sImage;
D3Vector3		sVector;
unsigned		uColor;
unsigned		uColorR;
unsigned		uColorG;
unsigned		uColorB;
double			dValue;
int				iPosY;
int				iPosX;
int				iMul;



	sField.SetSize(32,pData->iSteps,pData->iSteps);

	for(iPosY=0;iPosY<pData->iSteps;iPosY++)
	for(iPosX=0;iPosX<pData->iSteps;iPosX++)
		{
		dValue				= pValues[iPosY*pData->iSteps+iPosX];
		sVector.sVector.x	= PlaneStepToPos(pData,iPosX);
		sVector.sVector.y	= PlaneStepToPos(pData,iPosY);
		sVector.sVector.z	= PlaneValToPos (pData,dValue,2);

		if(sVector.sVector.z<-1.0)
			{
			if(sVector.sVector.z>-1.25)
					uColorR = (int)((sVector.sVector.z+1.0)*-1020.0);
			else	uColorR = 255;

			uColorB	= 255-uColorR/2;

			uColor  = uColorR<<16;
			uColor |= uColorB;
			}
		else if(sVector.sVector.z<=1.0)
			{
			uColorB	= (int)((1.0-sVector.sVector.z)*127.5);
			uColorG	= 255-uColorB;

			uColor  = uColorG<<8;
			uColor |= uColorB;
			}
		else{
			if(sVector.sVector.z<1.25)
					uColorR = (int)((sVector.sVector.z-1.0)*1020.0);
			else	uColorR = 255;

			uColorG = 255-uColorR;

			uColor  = uColorR<<16;
			uColor |= uColorG;
			}

		sField.SetPixel(iPosX,iPosY,uColor);
		}

								  iMul=1;
	if(pData->uOptions&PDR_DOUBLE)iMul=2;
	if(pData->uOptions&PDR_QUARD )iMul=4;

	if(iMul>1)
		{
		sImage.Stretch(sField,pData->iSteps*2,pData->iSteps*2);
		pImage->Put(sImage,iMul*FRAME_XY,iMul*FRAME_Z);
		}
	else{
		pImage->Put(sField,FRAME_XY,FRAME_Z);
		}

}

//*****************************************************************************
//*
//*		PlaneDrawColorBar
//*
//*****************************************************************************
//	Zeichnet eine Farbige Skala rechts oben
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	uColor		: Textfarbe
void PlaneDrawColorBar(PlaneData *pData,BitmapImage *pImage,unsigned uColor)
{
PlaneScala	sScala;
double		dSteps;
double		dValue;
double		dComma;
double		dFact;
double		dLog;
double		dMin;
double		dMax;
double		dVal;
char		cText[64];
unsigned	uColorB;
unsigned	uColorG;
int			iSteps;
int			iSizeY;
int			iPosX;
int			iPosY;
int			iBase;
int			iAdd;
int			iVal;
int			iSub;
int			iPos;
int			iLen;
int			iIdx;
int			iNum;


	if(!pImage->IsRGB())return;

	iSizeY	= pData->iSizeY/3-FRAME_Z;
	if(iSizeY<50)iSizeY=50;
	iSteps	= iSizeY/STEP_SIZE_ZP;
	iLen	= COLORBOX_WIDTH;
	iPosY	= FRAME_Z;

	if(pData->uOptions&PDR_DOUBLE){iSteps/=2;iLen*=2;iPosY*=2;}
	if(pData->uOptions&PDR_QUARD ){iSteps/=4;iLen*=4;iPosY*=4;}

	if(iSteps<1)iSteps=1;
	iPosX	= pData->iSizeX-((iPosY/FRAME_Z)*FRAME_XY)-iLen+1;
	iAdd	= -3;

	memset(&sScala,0,sizeof(sScala));
	sScala.dMax	= pData->aScala[2].dMax;
	sScala.dMin	= pData->aScala[2].dMin;


	if(pData->uOptions&(PDR_IS_LOG_Z))						// Logarithmisch berechnen
		{
	 	iAdd = PlaneCalcLog(&sScala,iSteps);
		}

	if(iAdd<=-3)											// Liear berechnen
		{
		iAdd = PlaneCalcLinear(&sScala,iSteps);
		}


	if(iAdd>=0)												// Lineare Anzeige
		{
		dSteps	= sScala.aSteps[0];
		dValue	= fabs(pData->aScala[2].dMin);
		dComma	= fabs(pData->aScala[2].dMax);
		if(dComma>dValue)dValue=dComma;


		iVal	= (int)floor(log10(dValue));				// Stellen vom maximal Wert
		iSub	= (int)floor(log10(dSteps))-iAdd;			// Stellen von der Schrittbreite
		iBase	= 8;

		while(iVal<-2 && iSub<-3 && iBase>0)				// Nachkommstellen reduzieren
			{
			iVal  += 3;
			iSub  += 3;
			iBase -= 1;
			}

		while(iVal>2 && iSub>3 && iBase<15)					// Vorkommstellen reduzieren
			{
			iVal  -= 3;
			iSub  -= 3;
			iBase += 1;
			}

		if(iVal<0)											// Wert hat nur Nachkommastellen
			{
			if(iSub>0)iSub=0;
			iVal  = 1-iSub;
			}
		else if(iSub<0)										// Wert hat Vor- und Nachkommastellen
			{
			iVal -= iSub-1;
			}
		else{
			iVal += 1;
			}

		if(iBase!=8)iVal++;
		if(iSub<0  )iVal++;
		else		iSub=0;


		if(pData->aScala[2].dMin<0.0)						// Gibt es ein Minuszeichen
			{
			iVal++;
			}

		sScala.iPlaces	= iVal;
		sScala.iComma	=-iSub;
		sScala.iBase	= iBase;
		}
	else{													// Logarithmische Anzeige
		sScala.iPlaces	= (iAdd==-2)? 6:5;
		sScala.iComma	= (iAdd==-2)? 1:0;
		sScala.iBase	= -1;
		}


	for(iPos=0;iPos<=iSizeY;iPos++)							// Farbige Skala
		{
		uColorB	=(255*iPos)/iSizeY;
		uColorG	= 255-uColorB;
		pImage->Box(iPosX,iPosY+iPos,iLen,1,(uColorG<<8)|uColorB);
		}


	if(pData->uOptions&PDR_QUARD)
		{
		pImage->SetFont(aFont3);
		iSub   = FONT_OFFSET_3;
		}
	else if(pData->uOptions&PDR_DOUBLE)
		{
		pImage->SetFont(aFont2);
		iSub   = FONT_OFFSET_2;
		}
	else{
		pImage->SetFont(aFont1);
		iSub   = FONT_OFFSET_1;
		}

	pImage->SetAlign(2);
	pImage->SetPixelFunction(TEXT_NONE,0,1);
	pImage->SetPixelFunction(TEXT_SMOTH,uColor,0);
	iPosX -= iPosY/3;
	iPosY += iSizeY;


	if(sScala.iIsLog)										// Logarithmische Teilung
		{
		dLog	= iSizeY;
		dLog   /= log(sScala.dMax/sScala.dMin);
		dVal	= sScala.dStart;
		dMax	= sScala.dMax*1.00001;
		dMin	= sScala.dMin*0.99999;
		iIdx	= sScala.iStepAt;

		for(iNum=0;dVal<=dMax && iNum<64;iNum++)
			{
			iPos = (int)(log(dVal/dMin)*dLog);

			if(iPos>=0 && iPos<=iSizeY)						// Beschriftung
				{
				pImage->Print(iPosX,iPosY-iPos-iSub,ScalaText(&sScala,cText,dVal));
				}

			dVal /= sScala.aSteps[iIdx];iIdx++;
			dVal *=	sScala.aSteps[iIdx];

			if(iIdx>=sScala.iStepCnt)iIdx=0;
			}
		}
	else{													// Lineare Teilung
		dVal	= sScala.dStart;
		dMax	= sScala.dMax+sScala.aSteps[0]*0.1;
		dMin	= sScala.dMin;
		dFact   = iSizeY;
		dFact  /=(sScala.dMax-sScala.dMin);

		for(iNum=0;dVal<=dMax && iNum<64;iNum++)
			{
			iPos = (int)((dVal-dMin)*dFact-0.1);

			if(iPos>=0 && iPos<=iSizeY)						// Beschriftung
				{
				pImage->Print(iPosX,iPosY-iPos-iSub,ScalaText(&sScala,cText,dVal));
				}

			dVal += sScala.aSteps[0];
			}
		}

}

//*****************************************************************************
//*
//*		PlaneInit
//*
//*****************************************************************************
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image
//	uOptions	: Optionen (diehe PDR_???)
//	pMinMax		: Zeiger auf die Min/Max-Werte der Scalen
//	pNames		: Name der Variablen und Einheiten (z.B. "A|A|B|B|C|C")
//	Ergibt die Anzahl der zu berechnenten Punkte
int PlaneInit(PlaneData *pData,BitmapImage *pImage,unsigned uOptions,double *pMinMax,const char *pNames,PlaneParams *pParams)
{
D3Matrix4x4		sOut;
D3Matrix4x4		sInv;
D3Matrix4x4		sSize;
D3Matrix4x4		sMove;
D3Matrix4x4		sProj;
D3Matrix4x4		sView;
D3Matrix4x4		sTemp;
D3Matrix4x4		sBase;
D3Matrix4x4		sRoll;
D3Vector3		sLook;
D3Vector3		sEye;
D3Vector3		sUp;
double			dComma;
double			dSteps;
double			dValue;
int				iGrid;
int				iSizeX;
int				iSizeY;
int				iScaleX;
int				iScaleY;
int				iScale;
int				iSteps;
int				iBase;
int				iMul;
int				iDiv;
int				iNum;
int				iVal;
int				iAdd;
int				iSub;
int				iLen;


	if(!pParams)pParams=&sDefParam;
	if(!pImage )return 1;

	iSizeX		=  pImage->Width ();
	iSizeY		=  pImage->Height();

	if(!pImage->IsRGB())
		{
		uOptions &= ~PDR_DOUBLE;
		uOptions &= ~PDR_QUARD;
		}

	if(uOptions&PDR_QUARD)								// Image Viermal so gro zeichnen
		{
		iDiv	  = 16;
		iMul	  =  4;
		iGrid     =  4;
		iSizeX   *=  4;
		iSizeY	 *=  4;
		uOptions |=  PDR_QUARD;
		uOptions &= ~PDR_DOUBLE;
		}
	else if(uOptions&PDR_DOUBLE)						// Image Doppelt so gro zeichnen
		{
		iDiv	  = 32;
		iMul	  =  2;
		iGrid     =  1;
		iSizeX   *=  2;
		iSizeY	 *=  2;
		uOptions |=  PDR_DOUBLE;
		uOptions &= ~PDR_QUARD;
		}
	else{												// Image eins zu eins zeichnen
		iDiv	  = 16;
		iMul	  =  1;
		iGrid     =  1;
		uOptions &= ~PDR_DOUBLE;
		uOptions &= ~PDR_QUARD;
		}

	if(uOptions&PDR_GRIDDOUBLE)
		{
		iGrid *= 2;
		}

	pData->iSizeX		= iSizeX;
	pData->iSizeY		= iSizeY;
	pData->uGridSub		= iGrid;
	pData->uGridMask	= iGrid-1;
	pData->dRollX		= pParams->dRollX;
	pData->dRollY		= pParams->dRollY;
	pData->dRollZ		= pParams->dRollZ;
	pData->dZoom		= pParams->dZoom;
	pData->dLook		= pParams->dLook;
	pData->dShadow		= pParams->dShadow;
	pData->iShadowPlane	= pParams->iShadowPlane;

	if(pData->dRollX<-45.0)pData->dRollX=-45.0;
	if(pData->dRollY<-45.0)pData->dRollY=-45.0;
	if(pData->dRollZ<-45.0)pData->dRollZ=-45.0;
	if(pData->dRollX>+45.0)pData->dRollX=+45.0;
	if(pData->dRollY>+45.0)pData->dRollY=+45.0;
	if(pData->dZoom < -1.0)pData->dZoom = -1.0;
	if(pData->dZoom > +1.0)pData->dZoom = +1.0;
	if(pData->dLook < -1.0)pData->dLook = -1.0;
	if(pData->dLook > +1.0)pData->dLook = +1.0;

	pData->dRollX  *= D3_PI/180.0;
	pData->dRollY  *= D3_PI/180.0;
	pData->dRollZ  *= D3_PI/180.0;
	pData->dZoom   *= -0.3;
	pData->dLook   *= 1.0;

	MatrixPerspectiveFovLH(&sProj,
						   D3_PI*0.22+pData->dZoom,		// Sichtwinkel der Kamera
						   1.0,							// Seitenverhltnis des Fensters
						   1.0,							// Minimale Entfernung frs Rendern
						   50.0); 						// Maximale Entfernung frs Rendern



	sBase.sMatrix.x1 =   0.0;							// PosX vom Auge
	sBase.sMatrix.x2 =  -8.0;							// PosY vom Auge
	sBase.sMatrix.x3 =   6.6;							// PosZ vom Auge
	sBase.sMatrix.x4 =   0.0;
	sBase.sMatrix.y1 =   0.0;							// LookX vom Auge
	sBase.sMatrix.y2 =   0.0;							// LookY vom Auge
	sBase.sMatrix.y3 =  -1.0+pData->dLook;				// LookZ vom Auge
	sBase.sMatrix.y4 =   0.0;
	sBase.sMatrix.z1 =   0.0;							// UP-Vektor
	sBase.sMatrix.z2 =   0.0;
	sBase.sMatrix.z3 =  -1.4;
	sBase.sMatrix.z4 =   0.0;
	sBase.sMatrix.w1 =   0.0;
	sBase.sMatrix.w2 =   0.0;
	sBase.sMatrix.w3 =   0.0;
	sBase.sMatrix.w4 =   0.0;



	MatrixRotationXYZ(&sRoll,pData->dRollX,pData->dRollY,pData->dRollZ+D3_PI);
	MatrixMultiply   (&sOut,&sBase,&sRoll);

	sEye.sVector.x  =  sOut.sMatrix.x1;
	sEye.sVector.y  = -sOut.sMatrix.x2;
	sEye.sVector.z  =  sOut.sMatrix.x3;

	sLook.sVector.x =  sOut.sMatrix.y1;
	sLook.sVector.y = -sOut.sMatrix.y2;
	sLook.sVector.z =  sOut.sMatrix.y3;

	sUp.sVector.x	=  sOut.sMatrix.z1;
	sUp.sVector.y	= -sOut.sMatrix.z2;
	sUp.sVector.z	=  sOut.sMatrix.z3;

	sEye.sVector.x +=  sLook.sVector.x;
	sEye.sVector.y +=  sLook.sVector.y;
	sEye.sVector.z +=  sLook.sVector.z;

	iScaleX			=  pData->iSizeX-FRAME_XY*2;
	iScaleY			=  pData->iSizeY-FRAME_Z *2;

	if(iScaleX<10)iScaleX=10;
	if(iScaleY<10)iScaleY=10;

	iScale			= (iScaleX<iScaleY)? iScaleX:iScaleY;
	pData->dFontScale = FONT_SCALE/(double)iScale;
															// Transformations Matrix erzeugen
	MatrixTranslation(&sMove,iScaleX/2,iScaleY/2,0.0);
	MatrixScaling    (&sSize,iScale,iScale, 1.0);
	MatrixLookAtLH   (&sView,&sEye ,&sLook,&sUp);
	MatrixMultiply   (&sTemp,&sView,&sProj);
	MatrixMultiply   (&sTemp,&sTemp,&sSize);
	MatrixMultiply   (&sTemp,&sTemp,&sMove);
	MatrixInversion	 (&sInv ,&sTemp);


	pData->sDiv		=  sInv;
	pData->sMul		=  sTemp;
	pData->iSteps	=  pData->iSizeX/iDiv;
	pData->iSteps  &=~(iGrid*iMul-1);
	pData->iSteps  |=  1;
	if(pData->iSteps<iGrid+1       )pData->iSteps=pData->uGridSub+1;
	if(pData->iSteps>GRID_MAX-iGrid)pData->iSteps=(GRID_MAX-iGrid)|1;
	pData->dStepMul =  2.0/(double)(pData->iSteps-1);


	if(!pNames)pNames="";


	for(iNum=0;iNum<3;iNum++)								// Werte fr Skalen zuweisen
		{
		pData->aScala[iNum].iIsLog		= uOptions&(PDR_IS_LOG_X<<iNum);
		pData->aScala[iNum].iMode		= iNum;
		pData->aScala[iNum].iDoubleMark	= 0;
		pData->aScala[iNum].iNameSize	= 0;
		pData->aScala[iNum].dMin		= pMinMax[iNum*2+0];
		pData->aScala[iNum].dMax		= pMinMax[iNum*2+1];

		iLen = strchar(pNames,'|');
		strlcpy(pData->aScala[iNum].cVarName,pNames,sizeof(pData->aScala[iNum].cVarName),iLen);
		pNames+=iLen;
		if(*pNames)pNames++;
		pData->aScala[iNum].iNameSize+=iLen+(iLen && *pNames!='|' && *pNames!=0)? 1:0;

		iLen = strchar(pNames,'|');
		strlcpy(pData->aScala[iNum].cUnitName,pNames,sizeof(pData->aScala[iNum].cUnitName),iLen);
		pNames+=iLen;
		if(*pNames)pNames++;
		pData->aScala[iNum].iNameSize+=iLen+(iLen)? 2:0;
		}

	pData->aScala[2].iDoubleMark	= 1;

	for(iNum=0;iNum<3;iNum++)								// Berechnung der Texte und Skalen
		{
		if(iNum<=1)											// X/Y Schritte abschtzen
			{
			iSteps	= (pData->iSizeX-FRAME_XY*2*iMul)/STEP_SIZE_XY;
			}
		else{												// Z Schritte abschtzen
			iSteps	= (pData->iSizeY-FRAME_Z*2*iMul)/STEP_SIZE_Z;
			}

		iSteps /= iMul;

		if(iSteps<1)iSteps=1;

		iAdd = -3;

		if(uOptions&(PDR_IS_LOG_X<<iNum))					// Logarithmisch berechnen
			{
			iAdd = PlaneCalcLog(pData->aScala+iNum,iSteps);
			}

		if(iAdd<=-3)										// Liear berechnen
			{
			iAdd = PlaneCalcLinear(pData->aScala+iNum,iSteps);
			}


		if(iAdd>=0)											// Lineare Anzeige
			{
			dSteps	= pData->aScala[iNum].aSteps[0];
			dValue	= fabs(pMinMax[iNum*2+0]);
			dComma	= fabs(pMinMax[iNum*2+1]);
			if(dComma>dValue)dValue=dComma;


			iVal	= (int)floor(log10(dValue));			// Stellen vom maximal Wert
			iSub	= (int)floor(log10(dSteps))-iAdd;		// Stellen von der Schrittbreite
			iBase	= 8;

			while(iVal<-2 && iSub<-3 && iBase>0)			// Nachkommstellen reduzieren
				{
				iVal  += 3;
				iSub  += 3;
				iBase -= 1;
				}

			while(iVal>2 && iSub>3 && iBase<15)				// Vorkommstellen reduzieren
				{
				iVal  -= 3;
				iSub  -= 3;
				iBase += 1;
				}

			if(iVal<0)										// Wert hat nur Nachkommastellen
				{
				if(iSub>0)iSub=0;
				iVal  = 1-iSub;
				}
			else if(iSub<0)									// Wert hat Vor- und Nachkommastellen
				{
				iVal -= iSub-1;
				}
			else{
				iVal += 1;
				}

			if(iBase!=8)iVal++;
			if(iSub<0  )iVal++;
			else		iSub=0;


			if(pMinMax[iNum*2+0]<0.0)						// Gibt es ein Minuszeichen
				{
				iVal++;
				}

			pData->aScala[iNum].iPlaces = iVal;
			pData->aScala[iNum].iComma	=-iSub;
			pData->aScala[iNum].iBase	= iBase;
			}
		else{												// Logarithmische Anzeige
			pData->aScala[iNum].iPlaces = (iAdd==-2)? 6:5;
			pData->aScala[iNum].iComma	= (iAdd==-2)? 1:0;
			pData->aScala[iNum].iBase	= -1;
			}
		}



	pData->uOptions	= uOptions;




return pData->iSteps;
}

//*****************************************************************************
//*
//*		PlaneDraw
//*
//*****************************************************************************
//	Zeichnet eine dreidimensionale Ebene
//	pData		: Zeiger auf die Kurvendaten
//	pImage		: Zeiger auf das Image in dem gezeichnet wird
//	pValues		: Werte fr die Z-Ebene auf
void PlaneDraw(PlaneData *pData,BitmapImage *pImage,double *pValues)
{
unsigned char  *pRow0;
unsigned char  *pRow1;
unsigned char  *pRow2;
unsigned char  *pRow3;
BitmapImage		sTemp;
BitmapImage    *pMain;
double			aSteps[GRID_MAX];
D3Vector3		sVectorA;
D3Vector3		sVectorB;
D3Vector3		sVectorC;
D3Vector3		sVectorN;
D3Vector3		sVectorW;
D3Vector3		sVector;
D3Vector3	   *pCoord;
double	 	   *pBuffer;
double	 	   *pNormal;
double			dStepP;
double			dStepN;
double			dValue;
unsigned		uSize;
unsigned		uColor;
unsigned		uColorR;
unsigned		uColorG;
unsigned		uColorB;
int				aArray[10];
int				iPosX;
int				iPosY;
int				iPos;
int				iMul;


	if(!pImage)return;

	if(pData->uOptions&(PDR_DOUBLE|PDR_QUARD))				// Doppelt/Viermal so gro zeichnen
		{
		pMain  =  pImage;
		pImage = &sTemp;
		sTemp.SetSize(32,pData->iSizeX,pData->iSizeY,COLOR_BACKPLANE);
		}
	else{
		pImage->Box(0,0,pImage->Width(),pImage->Height(),COLOR_BACKPLANE);
		}

								  iMul=1;
	if(pData->uOptions&PDR_DOUBLE)iMul=2;
	if(pData->uOptions&PDR_QUARD )iMul=4;

	   uSize = pData->iSizeX*pData->iSizeY;
	if(uSize>pData->uBufSizeZ)								// Z-Puffer vergrern
		{
		if(pData->uBufSizeZ)delete pData->pBufferZ;
		pBuffer	= new double [uSize];
		pData->uBufSizeZ = uSize;
		pData->pBufferZ  = pBuffer;
		}
	else{
		pBuffer = pData->pBufferZ;
		}


	pNormal	= new double   [GRID_MAX*pData->iSteps];
	pCoord	= new D3Vector3[GRID_MAX*pData->iSteps];

	dValue	= 1.0;
	dStepP	= pData->dStepMul;
	dStepN	=-pData->dStepMul;
	memsetqw(pBuffer,*((__int64*)&dValue),pData->iSizeX*pData->iSizeY);


	pData->pBufferZ		= pBuffer;							// Z-Buffer fr die Zeichnung
	pData->pGridValues	= pNormal;							// Normierte Gitter-Werte
	pData->pGridVects	= pCoord;							// Transformierte Gitter Vektoren
	pData->iMul			= iMul;								// Faktor fr Koordinaten
	pData->uWhite		= 0;
	pImage->SetUserPtr(pData);

	pData->sSunVect.sVector.x	=  0.0;
	pData->sSunVect.sVector.y	= -8.0;
	pData->sSunVect.sVector.z	= -8.0;
	VectorNormalize(&pData->sSunVect,&pData->sSunVect);

	if(pData->uOptions&PDR_COLORFIELD)						// Zeichen Farbenrechteck
		{
		PlaneDrawColorField(pData,pImage,pValues);
		}

	for(iPos=0;iPos<pData->iSteps;iPos++)					// Gitter Positionen
		{
		aSteps[iPos] = PlaneStepToPos(pData,iPos);
		}

	for(iPosY=0;iPosY<pData->iSteps;iPosY++)				// Gitter-Koordinaten berechnen
		{
		sVector.sVector.y		= aSteps[iPosY];

		for(iPosX=0;iPosX<pData->iSteps;iPosX++)
			{
			dValue				= pValues[iPosY*pData->iSteps+iPosX];
			sVector.sVector.x	= aSteps [iPosX];
			sVector.sVector.z	= PlaneValToPos(pData,dValue,2);
			pNormal[iPosY*GRID_MAX+iPosX] = sVector.sVector.z;

			VectorTransformCoord(pCoord+iPosY*GRID_MAX+iPosX,&pData->sMul,&sVector);
			}
		}

	if(pData->uOptions&PDR_GRID)							// Gitter zeichnen
		{
		PlaneDrawGrid(pData,pImage,0.0001);
		}


	if(pData->uOptions&PDR_PLANE)							// Ebene mit Dreiecken zeichnen
	for(iPosY=1;iPosY<pData->iSteps;iPosY++)
	for(iPosX=1;iPosX<pData->iSteps;iPosX++)
		{
		aArray[0]	= (int)pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.x;
		aArray[1]	= (int)pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.y;
		aArray[2]	= (int)pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		aArray[3]	= (int)pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;
		aArray[4]	= (int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.x;
		aArray[5]	= (int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.y;
		aArray[6]	= (int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.x;
		aArray[7]	= (int)pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.y;
		aArray[8]	= aArray[0];
		aArray[9]	= aArray[1];

		sVectorA.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.x;
		sVectorA.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		sVectorA.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.y;
		sVectorA.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;
		sVectorA.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.z;
		sVectorA.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.z;

		sVectorB.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.x;
		sVectorB.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.x;
		sVectorB.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.y;
		sVectorB.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.y;
		sVectorB.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.z;
		sVectorB.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX           ].sVector.z;

		VectorCross	   (&sVector ,&sVectorA,&sVectorB);
		pData->dAddZ	= pCoord[iPosY*GRID_MAX+iPosX].sVector.z;
		pData->dFactX	= sVector.sVector.x;
		pData->dFactX  /= sVector.sVector.z;
		pData->dFactY	= sVector.sVector.y;
		pData->dFactY  /= sVector.sVector.z;
		pData->iPosX	= aArray[2];
		pData->iPosY	= aArray[3];

		if(pData->uOptions&PDR_LIGHT)						// Licht berechnen
			{
			sVectorA.sVector.x  = dStepN;
			sVectorA.sVector.y  = 0.0;
			sVectorA.sVector.z  = pNormal[iPosY*GRID_MAX+iPosX-1];
			sVectorA.sVector.z -= pNormal[iPosY*GRID_MAX+iPosX  ];

			sVectorB.sVector.x  = 0.0;
			sVectorB.sVector.y  = dStepN;
			sVectorB.sVector.z  = pNormal[iPosY*GRID_MAX+iPosX-GRID_MAX];
			sVectorB.sVector.z -= pNormal[iPosY*GRID_MAX+iPosX         ];

			VectorCross    (&sVectorC,&sVectorA,&sVectorB);
			VectorNormalize(&sVectorN,&sVectorC);
			VectorCross	   (&sVectorW,&sVectorN,&pData->sSunVect);

			dValue = WHITE_MAX-VectorLength(&sVectorW)*(double)WHITE_MAX;

				 if(dValue>WHITE_MAX)pData->uWhite=WHITE_MAX;
			else if(dValue< 0.0     )pData->uWhite=0;
			else				     pData->uWhite=(int)dValue;
			}

		pImage->Triangle(aArray,BGR(0,255,0),SetTriangel);

		sVectorA.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.x;
		sVectorA.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.x;
		sVectorA.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.y;
		sVectorA.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.y;
		sVectorA.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-1         ].sVector.z;
		sVectorA.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.z;

		sVectorB.sVector.x  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.x;
		sVectorB.sVector.x -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.x;
		sVectorB.sVector.y  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.y;
		sVectorB.sVector.y -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.y;
		sVectorB.sVector.z  = pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX  ].sVector.z;
		sVectorB.sVector.z -= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.z;

		VectorCross    (&sVector ,&sVectorA,&sVectorB);

		pData->dAddZ	= pCoord[iPosY*GRID_MAX+iPosX-GRID_MAX-1].sVector.z;
		pData->dFactX	= sVector.sVector.x;
		pData->dFactX  /= sVector.sVector.z;
		pData->dFactY	= sVector.sVector.y;
		pData->dFactY  /= sVector.sVector.z;
		pData->iPosX	= aArray[6];
		pData->iPosY	= aArray[7];

		if(pData->uOptions&PDR_LIGHT)						// Licht berechnen
			{
			sVectorA.sVector.x  = 0.0;
			sVectorA.sVector.y  = dStepP;
			sVectorA.sVector.z  = pNormal[iPosY*GRID_MAX+iPosX-1         ];
			sVectorA.sVector.z -= pNormal[iPosY*GRID_MAX+iPosX-GRID_MAX-1];

			sVectorB.sVector.x  = dStepP;
			sVectorB.sVector.y  = 0.0;
			sVectorB.sVector.z  = pNormal[iPosY*GRID_MAX+iPosX-GRID_MAX  ];
			sVectorB.sVector.z -= pNormal[iPosY*GRID_MAX+iPosX-GRID_MAX-1];

			VectorCross    (&sVectorC,&sVectorA,&sVectorB);
			VectorNormalize(&sVectorN,&sVectorC);
			VectorCross	   (&sVectorW,&sVectorN,&pData->sSunVect);

			dValue = WHITE_MAX-VectorLength(&sVectorW)*(double)WHITE_MAX;

				 if(dValue>WHITE_MAX)pData->uWhite=WHITE_MAX;
			else if(dValue< 0.0     )pData->uWhite=0;
			else				     pData->uWhite=(int)dValue;
			}

		pImage->Triangle(aArray+4,BGR(0,0,255),SetTriangel);
		}



	if(pData->uOptions&PDR_SCALA_X)PlaneDrawScala(pData,pImage,pData->aScala+0,COLOR_SCALA_TEXT_X);
	if(pData->uOptions&PDR_SCALA_Y)PlaneDrawScala(pData,pImage,pData->aScala+1,COLOR_SCALA_TEXT_Y);
	if(pData->uOptions&PDR_SCALA_Z)PlaneDrawScala(pData,pImage,pData->aScala+2,COLOR_SCALA_TEXT_Z);


	if(pData->uOptions&PDR_GRID)							// Gitter zeichnen
		{
		PlaneDrawGrid(pData,pImage,-0.0001);
		}

	if(pData->uOptions&PDR_SHADOW)							// Schatten zeichnen
		{
		PlaneDrawShadow(pData,pImage,pData->dShadow,pData->iShadowPlane);
		}

	if(pData->uOptions&PDR_COLORBAR)						// Farbskala anzeigen
		{
		PlaneDrawColorBar(pData,pImage,COLOR_SCALA_TEXT_Z);
		}


	if(pData->uOptions&PDR_DOUBLE)							// Doppelt so gro zeichnen
		{
		int		iSizeX;
		int		iSizeY;
		int		iPosX;
		int		iPosY;

		iSizeX = sTemp.Width ()/2;
		iSizeY = sTemp.Height()/2;

		for(iPosY=0;iPosY<iSizeY;iPosY++)
			{
			pRow0 = sTemp.GetRow4(iPosY*2+0);
			pRow1 = sTemp.GetRow4(iPosY*2+1);

			for(iPosX=0;iPosX<iSizeX;iPosX++)
				{
				uColor  = ((pRow0[iPosX*8+0]+pRow0[iPosX*8+4]+pRow1[iPosX*8+0]+pRow1[iPosX*8+4])/4);
				uColor |= ((pRow0[iPosX*8+1]+pRow0[iPosX*8+5]+pRow1[iPosX*8+1]+pRow1[iPosX*8+5])/4)<<8;
				uColor |= ((pRow0[iPosX*8+2]+pRow0[iPosX*8+6]+pRow1[iPosX*8+2]+pRow1[iPosX*8+6])/4)<<16;

				pMain->SetPixel(iPosX,iPosY,uColor);
				}
			}
		}

	if(pData->uOptions&PDR_QUARD)							// Viermal so gro zeichnen
		{
		int		iSizeX;
		int		iSizeY;
		int		iPosX;
		int		iPosY;


		iSizeX = sTemp.Width ()/4;
		iSizeY = sTemp.Height()/4;

		for(iPosY=0;iPosY<iSizeY;iPosY++)
			{
			pRow0 = sTemp.GetRow4(iPosY*4+0);
			pRow1 = sTemp.GetRow4(iPosY*4+1);
			pRow2 = sTemp.GetRow4(iPosY*4+2);
			pRow3 = sTemp.GetRow4(iPosY*4+3);

			for(iPosX=0;iPosX<iSizeX;iPosX++)
				{
				uColorB  = pRow0[iPosX*16+0]+pRow0[iPosX*16+4]+pRow0[iPosX*16+ 8]+pRow0[iPosX*16+12];
				uColorG  = pRow0[iPosX*16+1]+pRow0[iPosX*16+5]+pRow0[iPosX*16+ 9]+pRow0[iPosX*16+13];
				uColorR  = pRow0[iPosX*16+2]+pRow0[iPosX*16+6]+pRow0[iPosX*16+10]+pRow0[iPosX*16+14];

				uColorB += pRow1[iPosX*16+0]+pRow1[iPosX*16+4]+pRow1[iPosX*16+ 8]+pRow1[iPosX*16+12];
				uColorG += pRow1[iPosX*16+1]+pRow1[iPosX*16+5]+pRow1[iPosX*16+ 9]+pRow1[iPosX*16+13];
				uColorR += pRow1[iPosX*16+2]+pRow1[iPosX*16+6]+pRow1[iPosX*16+10]+pRow1[iPosX*16+14];

				uColorB += pRow2[iPosX*16+0]+pRow2[iPosX*16+4]+pRow2[iPosX*16+ 8]+pRow2[iPosX*16+12];
				uColorG += pRow2[iPosX*16+1]+pRow2[iPosX*16+5]+pRow2[iPosX*16+ 9]+pRow2[iPosX*16+13];
				uColorR += pRow2[iPosX*16+2]+pRow2[iPosX*16+6]+pRow2[iPosX*16+10]+pRow2[iPosX*16+14];

				uColorB += pRow3[iPosX*16+0]+pRow3[iPosX*16+4]+pRow3[iPosX*16+ 8]+pRow3[iPosX*16+12];
				uColorG += pRow3[iPosX*16+1]+pRow3[iPosX*16+5]+pRow3[iPosX*16+ 9]+pRow3[iPosX*16+13];
				uColorR += pRow3[iPosX*16+2]+pRow3[iPosX*16+6]+pRow3[iPosX*16+10]+pRow3[iPosX*16+14];

				uColor   = (uColorB/16);
				uColor  |= (uColorG/16)<<8;
				uColor  |= (uColorR/16)<<16;

				pMain->SetPixel(iPosX,iPosY,uColor);
				}
			}
		}



	delete pNormal;
	delete pCoord;


}


//*****************************************************************************
//*
//*		PlaneCoorToVal
//*
//*****************************************************************************
//	Rechnet Koordinaten vom Bild in die Werte der Ebene um.
//	pData		: Zeiger auf die Kurvendaten
//	iPosX/Y		: Koordinaten im Bild
//	pValues		: Hier werden die Werte der Ebene gespeichert
//	Ergibt 1 wenn der Punkt auf der Ebene ist.
int PlaneCoorToVal(PlaneData *pData,int iPosX,int iPosY,double *pValues)
{
D3Vector3	sVectorA;
D3Vector3	sVectorB;
double		dVal;
int			iPos;


	iPosX *= pData->iMul;
	iPosY *= pData->iMul;


	if((unsigned)iPosX>=(unsigned)pData->iSizeX)
		{
		pValues[0] = 0;
		pValues[1] = 0;
		pValues[2] = 0;

		return 0;
		}

	if((unsigned)iPosY>=(unsigned)pData->iSizeY)
		{
		pValues[0] = 0;
		pValues[1] = 0;
		pValues[2] = 0;

		return 0;
		}


	iPos = iPosX+iPosY*pData->iSizeX;
	dVal = pData->pBufferZ[iPos];

	if(dVal>=1.0 || dVal<=0.0)
		{
		pValues[0] = 0;
		pValues[1] = 0;
		pValues[2] = 0;

		return 0;
		}

	sVectorA.sVector.x	= iPosX;
	sVectorA.sVector.y	= iPosY;
	sVectorA.sVector.z	= dVal;

	VectorTransformCoord(&sVectorB,&pData->sDiv,&sVectorA);


	if((sVectorB.sVector.x<-0.993 && sVectorB.sVector.y<-0.993) ||
	   (sVectorB.sVector.x<-1.002 || sVectorB.sVector.y<-1.002))
		{
		pValues[0] = 0;
		pValues[1] = 0;
		pValues[2] = 0;

		return 0;
		}

	pValues[0] = PlanePosToVal(pData,sVectorB.sVector.x,0);
	pValues[1] = PlanePosToVal(pData,sVectorB.sVector.y,1);
	pValues[2] = PlanePosToVal(pData,sVectorB.sVector.z,2);




return 1;
}


//*****************************************************************************
//*
//*		PlaneCoorToVal
//*
//*****************************************************************************
//	Rechnet Werte in der Ebene in Koordinaten vom Bild um.
//	pData		: Zeiger auf die Kurvendaten
//	iPosX/Y		: Hier werden die Koordinaten im Bild abgelegt
//	pValues		: Werte der Ebene
//	Ergibt 1 wenn wenn der Punkt im Wrfel ist
int PlaneValToCoor(PlaneData *pData,int&iPosX,int&iPosY,double *pValues)
{
D3Vector3	sVectorA;
D3Vector3	sVectorB;
int			iOk;



	sVectorA.sVector.x = PlaneValToPos(pData,pValues[0],0);
	sVectorA.sVector.y = PlaneValToPos(pData,pValues[1],1);
	sVectorA.sVector.z = PlaneValToPos(pData,pValues[2],2);

	if(sVectorA.sVector.x>1.00001 || sVectorA.sVector.x<-1.00001)iOk=0;
	if(sVectorA.sVector.y>1.00001 || sVectorA.sVector.y<-1.00001)iOk=0;
	if(sVectorA.sVector.z>1.00001 || sVectorA.sVector.z<-1.00001)iOk=0;

	VectorTransformCoord(&sVectorB,&pData->sMul,&sVectorA);

	iOk = 1;


	iPosX = (int)sVectorB.sVector.x/pData->iMul;
	iPosY = (int)sVectorB.sVector.y/pData->iMul;


return iOk;
}

